const std = @import("std");
const mem = std.mem;
const heap = std.heap;

const h = @import("libvirt-helper.zig");
const c = @import("libvirt-c.zig").c;
const err = @import("libvirt-error.zig");
const Domain = @import("libvirt-domain.zig");
const Pool = @import("libvirt-pool.zig");
const VirError = err.VirError;
const Connection = @This();

ptr: c.virConnectPtr,
allocator: mem.Allocator,

pub fn init(uri: []const u8, allocator: mem.Allocator) VirError!Connection {
    const connection = c.virConnectOpenAuth(@ptrCast(uri), c.virConnectAuthPtrDefault, 0);
    if (connection) |conn| return .{
        .ptr = conn,
        .allocator = allocator,
    } else return err.handleError();
}

pub fn deinit(self: *const Connection) void {
    _ = c.virConnectClose(self.ptr);
}

pub fn getURI(self: *const Connection) Error![]u8 {
    const uri = c.virConnectGetURI(self.ptr);
    defer std.c.free(uri);
    const str = h.string(uri);
    return if (str.len == 0) err.handleError() else try self.allocator.dupe(u8, str);
}
pub fn freeURI(self: *const Connection, uri: []u8) void {
    self.allocator.free(uri);
}

test "connection" {
    const conn = try Connection.init("qemu:///system", std.testing.allocator);
    defer conn.deinit();

    const uri = try conn.getURI();
    defer conn.freeURI(uri);
}

// pub fn createPool(self: *const Connection, xml: String, flags: []const Pool.CreateFlags) !Pool {
//     const pool = c.virStoragePoolCreateXML(self.ptr, @ptrCast(xml.str), intFromFlags(Pool.CreateFlags, flags));
//     return if (pool) |p| Pool.init(p, self.allocator) else handleError();
// }
// pub fn definePoolFlags(self: *const Connection, xml: String, flags: []const Pool.DefineFlags) !Pool {
//     const pool = c.virStoragePoolCreateXML(self.ptr, @ptrCast(xml.str), intFromFlags(Pool.DefineFlags, flags));
//     return if (pool) |p| Pool.init(p, self.allocator) else handleError();
// }

pub const Error = mem.Allocator.Error || VirError;

pub fn pools(self: *const Connection) Error![]Pool {
    const list = try Pool.listAllStoragePools(self, &[_]Pool.ListFlags{});
    var l = try self.allocator.alloc(Pool, list.len);
    for (list, 0..) |p, i| l[i] = Pool.init(p, self.allocator);
    return l;
}
pub fn freePools(self: *const Connection, slice: []Pool) void {
    // for (slice) |pool| pool.deinit();
    self.allocator.free(slice);
}
pub fn iteratePools(self: *const Connection) Error!PoolIterator {
    return .{
        .allocator = self.allocator,
        .values = try self.pools(),
    };
}
pub const PoolIterator = struct {
    allocator: mem.Allocator,
    values: []Pool,
    index: usize = 0,

    pub fn deinit(it: *PoolIterator) void {
        for (it.values) |v| v.free() catch @panic("pool free error");
        it.allocator.free(it.values);
    }
    pub fn next(it: *PoolIterator) ?Pool {
        if (it.index >= it.values.len) return null;
        const v = it.values[it.index];
        it.index += 1;
        return v;
    }
    pub fn reset(it: *PoolIterator) void {
        it.index = 0;
    }
};

pub fn numOfDomains(self: *const Connection) VirError!u32 {
    return Domain.numOfDomains(self);
}
pub fn numOfDefinedDomains(self: *const Connection) VirError!u32 {
    return Domain.numOfDefinedDomains(self);
}

// pub fn createDomain(self: *const Connection, xml: String, flags: []const Domain.CreateFlags) !Domain {
//     const domain = c.virDomainCreateXML(self.ptr, @ptrCast(xml.str), intFromFlags(Domain.CreateFlags, flags));
//     return if (domain) |d| Domain.init(d, self.allocator) else handleError();
// }
// pub fn defineDomain(self: *const Connection, xml: String) !Domain {
//     const domain = c.virDomainDefineXML(self.ptr, @ptrCast(xml.str));
//     return if (domain) |d| Domain.init(d, self.allocator) else handleError();
// }
// pub fn defineDomainFlags(self: *const Connection, xml: String, flags: []const Domain.DefineFlags) !Domain {
//     const domain = c.virDomainDefineXMLFlags(self.ptr, @ptrCast(xml.str), intFromFlags(Domain.DefineFlags, flags));
//     return if (domain) |d| Domain.init(d, self.allocator) else handleError();
// }

pub fn createDomain(self: *const Connection, xml: h.String, flags: []const Domain.CreateFlags) VirError!Domain {
    const ptr = try Domain.createXML(self, xml, flags);
    return Domain.init(ptr, self.allocator);
}

pub fn domains(self: *const Connection) Error![]Domain {
    const list = try Domain.listAllDomains(self, &[_]Domain.ListFlags{});
    var l = try self.allocator.alloc(Domain, list.len);
    for (list, 0..) |p, i| l[i] = Domain.init(p, self.allocator);
    return l;
}
pub fn freeDomains(self: *const Connection, slice: []Domain) void {
    // for (slice) |domain| domain.deinit();
    self.allocator.free(slice);
}
pub fn iterateDomains(self: *const Connection) Error!DomainIterator {
    return .{
        .allocator = self.allocator,
        .values = try self.domains(),
    };
}
pub const DomainIterator = struct {
    allocator: mem.Allocator,
    values: []Domain,
    index: usize = 0,

    pub fn deinit(it: *DomainIterator) void {
        for (it.values) |v| v.free() catch @panic("domain free error");
        it.allocator.free(it.values);
    }
    pub fn next(it: *DomainIterator) ?Domain {
        if (it.index >= it.values.len) return null;
        const v = it.values[it.index];
        it.index += 1;
        return v;
    }
    pub fn reset(it: *DomainIterator) void {
        it.index = 0;
    }
};

test "domains" {
    const conn = try Connection.init("qemu:///system", std.testing.allocator);
    defer conn.deinit();

    // Enumerate domains
    const num_active = try conn.numOfDomains();
    const num_inactive = try conn.numOfDefinedDomains();
    _ = .{ num_active, num_inactive };

    // Iterate domains
    var domain_it = try conn.iterateDomains();
    defer domain_it.deinit();
    while (domain_it.next()) |domain| {
        const active = try domain.isActive();
        const name = try domain.getName();
        _ = .{ active, name };
    }

    // Create and destory domain
    const xml = try h.String.init(std.testing.allocator,
        \\<domain type="kvm">
        \\  <name>vm1</name>
        \\  <uuid>48dd0909-6346-4028-b922-1b3a13571360</uuid>
        \\  <memory unit="KiB">1048576</memory>
        \\  <currentMemory unit="KiB">1048576</currentMemory>
        \\  <vcpu placement="static">1</vcpu>
        \\  <os>
        \\    <type arch="x86_64" machine="pc-i440fx-8.2">hvm</type>
        \\    <boot dev="hd"/>
        \\  </os>
        \\  <features>
        \\    <acpi/>
        \\    <apic/>
        \\    <vmport state="off"/>
        \\  </features>
        \\  <cpu mode="host-passthrough" check="none" migratable="on"/>
        \\  <clock offset="utc">
        \\    <timer name="rtc" tickpolicy="catchup"/>
        \\    <timer name="pit" tickpolicy="delay"/>
        \\    <timer name="hpet" present="no"/>
        \\  </clock>
        \\  <on_poweroff>destroy</on_poweroff>
        \\  <on_reboot>restart</on_reboot>
        \\  <on_crash>destroy</on_crash>
        \\  <pm>
        \\    <suspend-to-mem enabled="no"/>
        \\    <suspend-to-disk enabled="no"/>
        \\  </pm>
        \\  <devices>
        \\    <emulator>/run/libvirt/nix-emulators/qemu-system-x86_64</emulator>
        \\    <controller type="usb" index="0" model="ich9-ehci1">
        \\      <address type="pci" domain="0x0000" bus="0x00" slot="0x02" function="0x7"/>
        \\    </controller>
        \\    <controller type="usb" index="0" model="ich9-uhci1">
        \\      <master startport="0"/>
        \\      <address type="pci" domain="0x0000" bus="0x00" slot="0x02" function="0x0" multifunction="on"/>
        \\    </controller>
        \\    <controller type="usb" index="0" model="ich9-uhci2">
        \\      <master startport="2"/>
        \\      <address type="pci" domain="0x0000" bus="0x00" slot="0x02" function="0x1"/>
        \\    </controller>
        \\    <controller type="usb" index="0" model="ich9-uhci3">
        \\      <master startport="4"/>
        \\      <address type="pci" domain="0x0000" bus="0x00" slot="0x02" function="0x2"/>
        \\    </controller>
        \\    <controller type="pci" index="0" model="pci-root"/>
        \\    <serial type="pty">
        \\      <target type="isa-serial" port="0">
        \\        <model name="isa-serial"/>
        \\      </target>
        \\    </serial>
        \\    <console type="pty">
        \\      <target type="serial" port="0"/>
        \\    </console>
        \\    <input type="mouse" bus="ps2"/>
        \\    <input type="keyboard" bus="ps2"/>
        \\    <audio id="1" type="none"/>
        \\    <memballoon model="virtio">
        \\      <address type="pci" domain="0x0000" bus="0x00" slot="0x03" function="0x0"/>
        \\    </memballoon>
        \\  </devices>
        \\</domain>
    );
    defer xml.deinit();
    const domain = try conn.createDomain(xml, &.{});
    try domain.destroy();
    // TODO with flags
}
